// C Source File
// Created 8/7/2010; 10:31:23 PM

// Token file reader for Solar89
// Reads in a token file and returns an array of tokenInfo objects

#include <tigcclib.h>

#include "tokeninfo.h"

//Removes the first character of a string
void removeFirstChar(char *string)
{
	int i;
	int length = strlen(string) - 1;

	char backupchar;
	char *backup = string;

	//Loop over the characters of the pointer
	for (i = 0; i < length; ++i)
	{
		backupchar = backup[i+1];
		*(string++) = backupchar;
	}
	
	//Add a \0 in place of the last character
	*(string++) = '\0';
}

//Creates and appends a token info object to the tokens array
void createTokenInfo(tokenInfo *tokens, char *working, char *line, char *tokenstr, int i)
{
	int newlength;

	unsigned long rawhex;
	unsigned char hex = '0';
	
	tokenInfo token;

	//Init the token
	token = tokens[i];
	token.twoByteToken = 0;

	//If the line starts with a space, implement bugfix
	if (line[0] == ' ')
		removeFirstChar(line);

	//Seperate the name component from the line and set it (and it's length!)
	memset(working, '\0', strlen(working));
	sprintf(working, "%c", '\\');
	newlength = strcspn(line, working);
	tokenstr = strncat(tokenstr, line, newlength);
	line += (newlength + 1);
	token.chars = newlength;
	strncat(token.name, tokenstr, newlength);

	//Seperate the hex component from the line and set it
	memset(tokenstr, '\0', strlen(tokenstr));
	tokenstr = strncat(tokenstr, line, 4);
	rawhex = strtol(tokenstr, NULL, 0);
	hex = (unsigned char)rawhex;
	token.hex[0] = hex;
	line += 5;

	//Then check if what's left of the string is the second byte of a two-byte token and deal with it
	if (strcmp(line, "0") != 0)
	{
		rawhex = strtol(line, NULL, 0);
		hex = (unsigned char)rawhex;
		token.hex[1] = hex;
		token.twoByteToken = 1;
	}

	//Update the token
	tokens[i] = token;
}

//Initializes all token infos, return number of tokens
int initTokens(tokenInfo *tokens, char *tokentext, int NUM_TOKENS)
{
	int length;
	int span = 0;
	int i = 0;

	char newchar;
	char endline = '\r';

	char *working = NULL;
	char *line = NULL;
	char *tokenstr = NULL;

	if ((working = (char *)calloc(5, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for working string", BT_OK, BT_NONE);
		return -1;
	}
	if ((tokenstr = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token working string", BT_OK, BT_NONE);
		free(working);
		return -1;
	}
	if ((line = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for line string", BT_OK, BT_NONE);
		free(working);
		free(tokenstr);
		return -1;
	}

	//Get the length of the token file and realloc tokenstr (not sure if needed...)
	length = strlen(tokentext);

	//Loop until the length of the input string is zero
	while (length > 0)
	{
		//Reset the strings
		line = memset(line, '\0', strlen(line));
		tokenstr = memset(tokenstr, '\0', strlen(tokenstr));
		working = memset(working, '\0', strlen(working));

		//Get the length of the line up to trailing character, realloc mem
		sprintf(working, "%c", endline);
		span = strcspn(tokentext, working);
		if (tokenstr != NULL) 
			tokenstr = (char *)realloc(tokenstr, span * sizeof(char));
		if (line != NULL)
			line = (char *)realloc(line, span * sizeof(char));

		//Load the line into a string, and then remove it
		line = strncat(line, tokentext, span);
		tokentext = strpbrk(tokentext, working);
		
		//Then clear out the \r at the front of the string
		newchar = tokentext[2];
		sprintf(working, "%c", newchar);
		tokentext = strpbrk(tokentext, working);
		length = strlen(tokentext);

		//Then make a token info using some of these variables (to make this code easier to read, debug, and to fix a bug) and free the line memory
		createTokenInfo(tokens, working, line, tokenstr, i);
		i += 1;
	}
	
	//Free memory
	free(line);
	free(tokenstr);
	free(working);
	
	//Return the data
	return NUM_TOKENS;
}

//Reads a token file structure and token info array directly from an appvar, returns NUM_TOKENS
int readBinaryTokens(tokenInfo *tokens, char *filename, char *extension, int NUM_TOKENS)
{
	FILE *source = NULL;
	
	tokenFile tokenfile;

	//Open the input file and get its data
	if ((source = fopen(filename, "rb")) == NULL)
	{
		DlgMessage("Error", "Unable to open binary file", BT_OK, BT_NONE);
		return 0;
	}

	//In theory we should check that fread() works, but for some reason the check doesn't work

	//Now, read the beginning of the file into a token info structure
	fread(&tokenfile, 1, sizeof(tokenFile), source);

	//If the number of tokens is greater than the estimate, realloc the token array
	if (tokenfile.numtokens > NUM_TOKENS)
		realloc(tokens, tokenfile.numtokens * sizeof(tokenInfo));
		
	//Now, read into the token array
	fread(tokens, sizeof(tokenInfo) * NUM_TOKENS, 1, source);
	
	//Load the extension from the file into *extension
	extension = strcpy(extension, tokenfile.type);
	
	//Close the file and return
	fclose(source);
	return tokenfile.numtokens;
}