// C Source File
// Created 7/16/2010; 10:35:24 AM

// Solar89: A 68k C program to tokenize text files into 8XP program variables
// These variables can then be transferred to a computer for conversion
// To convert, install and use the Python language program make8xp

// Defined Comments (for shells)
#define COMMENT_STRING         "Tokenize z80 BASIC prgms"
#define COMMENT_PROGRAM_NAME   "Solar89"
#define COMMENT_VERSION_NUMBER 0,2,8,0
#define COMMENT_AUTHORS        "TC01"
#define COMMENT_BW_ICON \
	{0b0000111111110000, \
	 0b0000000011100000, \
	 0b0000000110000000, \
	 0b0000001100000000, \
	 0b0000111000000000, \
	 0b0001111111111000, \
	 0b0000000000000000, \
	 0b0111110001111100, \
	 0b1100011011000110, \
	 0b1100011011000110, \
	 0b1111111011111110, \
	 0b1111111011111110, \
	 0b1100011000000110, \
	 0b1100011000000110, \
	 0b0111110000000110, \
	 0b0000000000000000}

//Include solar89 header, which contains all GCC4TI libraries and functions called from external files
#include "solar89.h"

//Makes a dialog to get input strings, returns boolean (integer) of 0 or 1 depending on success
int getDialogInput(char *string)
{
	int result = 0;
	
	HANDLE dlg = H_NULL;

	//Create a dialog box and add the necessary inputs
	if ((dlg = DialogNewSimple(140, 55)) == H_NULL) 
	{
		DlgMessage("DMA Failure", "Unable to create dialog input box", BT_OK, BT_CANCEL);
		return result;
	}
	DialogAddTitle(dlg, "Solar89 Tokenizer", BT_OK, BT_CANCEL);
	DialogAddRequest(dlg, 5, 15, "Text File:", 0, 17, 18);
	DialogAddRequest(dlg, 5, 25, "Output File:", 17, 17, 18);
	DialogAddRequest(dlg, 5, 35, "Token File:", 34, 17, 18);
	
	//Prompt for input
	if (DialogDo(dlg, CENTER, CENTER, string, NULL) == KEY_ENTER)
		result = 1;

	HeapFree(dlg);
	
	//Return result to say we've done it
	return result;
}

//Extracts text from file, returns boolean
int getTextFromFile(char *filename, char *input, unsigned short size)
{
	FILE *source = NULL;

	int result = 0;
	
	//Open the input file and get its text
	if ((source = fopen(filename, "rt")) == NULL)
	{
		DlgMessage("I/O Error", "Unable to read output file- this may be because the file name was entered incorrectly.", BT_OK, BT_NONE);
		return result;
	}

	//In theory we should check that fread() works, but for some reason the check doesn't work
	/*if (fread(input, size, 1, source) != 1)
	{
		DlgMessage("Read Failure", "Unable to read input file", BT_OK, BT_NONE);
		return result;
	}*/
	
	result = 1;
	fread(input, size, 1, source);
	fclose(source);
	
	return result;
}

int writePrgmToVar(char *filename, unsigned char *output, unsigned short size, char *extension)
{
	unsigned i;
	unsigned length = 0;

	FILE *result = NULL;

	ST_helpMsg("Writing file");

	//Open a new file for writing in binary mode
	if ((result = fopen(filename, "wb")) == NULL)
	{
		DlgMessage("I/O Error", "Unable to create output file- this may be because that file already exists on your calc.", BT_OK, BT_NONE);
		return 0;
	}
	
	//Calculate length of output file
	for (i = 0; i < size; ++i)
	{
		if (output[i] == '\0' && output[i+1] == '\0')
			break;
		length += 1;
	}

	//Write all tokens in the output to file
	fwrite(output, sizeof(unsigned char), length, result);

	//Write the file type
	fputc(0, result);
	fputs(extension, result);
	fputc(0, result);
	fputc(OTH_TAG, result);

	//Close the file, archive it, and return 1
	fclose(result);
	EM_moveSymToExtMem(SYMSTR(filename), HS_NULL);
	return 1;
}

// Main Function
void _main(void)
{
	char *string = NULL;
	char *inputfile = NULL;
	char *outputfile = NULL;
	char *tokenfile = NULL;
	char *input = NULL;
	char *tokentext = NULL;
	char extension[5];

	unsigned char *output = NULL;
	unsigned short size = 0;

	int bool = 0;
	int length = 0;
	int numtokens;
	int NUM_TOKENS = 0;
	int argtype;
	int i = 0;
	int tokcomp = 0;
	
	SYM_ENTRY *sym;
	SYM_ENTRY *toksym;
	ESI argptr;
	ESQ type;
	ESQ inputtype;
	tokenInfo *tokens = NULL;

  //Allocate memory to pointers
	if ((string = (char *)calloc(60, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for raw input string", BT_OK, BT_NONE);
		return;
	}
	if ((inputfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for input file string", BT_OK, BT_NONE);
		free(string);
		return;
	}
	if ((outputfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		return;
	}
	if ((tokenfile = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("DMA Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		return;
	}
	
	//Check for arguments passed to the program
	if (ArgCount() == 3)
	{
		InitArgPtr(argptr);
		while ((argtype = GetArgType(argptr)) != END_TAG)
		{
			i += 1;
			if (argtype == STR_TAG)
			{
				switch (i)
				{
					case 1: 
						inputfile = strcpy(inputfile, GetStrnArg(argptr));
						continue;
					case 2: 
						outputfile = strcpy(outputfile, GetStrnArg(argptr));
						continue;
					case 3: 
						tokenfile = strcpy(tokenfile, GetStrnArg(argptr));
						continue;
					default: 
						continue;
				}
			}
			else
			{
				free(string);
				free(inputfile);
				free(outputfile);
				free(tokenfile);
				ER_throw(ER_ARGUMENT);
			}
		}
	}
	//If there are NOT three arguments, get dialog input, if this is impossible, free and quit
	else
	{
		bool = getDialogInput(string);
		if (bool == 0)
		{
			free(string);
			free(inputfile);
			free(outputfile);
			free(tokenfile);
			return;
		}

		//Retrieve file names from raw input
		for (length = 0; length < 17; ++length)
		{
			inputfile[length] = string[length];
			outputfile[length] = string[length + 17];
			tokenfile[length] = string[length + 34];
		}
	}
	
	//Find the size of the input and allocate memory accordingly
	ST_helpMsg("Reading input file");
	sym = SymFindPtr(SYMSTR(inputfile), 0);
	size = ((MULTI_EXPR*)HeapDeref(sym->handle))->Size + 2;
	inputtype = *(HToESI(sym->handle));
	if (224 != (int)inputtype)
	{
		DlgMessage("Invalid Input", "Program text input is not a text variable", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		return;
	}
	if ((input = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for input file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		return;
	}

	//Get the raw text of the text file
	bool = getTextFromFile(inputfile, input, size);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		return;	
	}
	
	//Allocate memory to the output array, based on input size
	if ((output = (unsigned char *)calloc(size, sizeof(unsigned char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for output file string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		return;
	}

	//Get the size and type of the token file- if the type is incorrect, quit
	ST_helpMsg("Initializing tokens");
	toksym = SymFindPtr(SYMSTR(tokenfile), 0);
	size = ((MULTI_EXPR*)HeapDeref(toksym->handle))->Size + 2;
	type = *(HToESI(toksym->handle));
	NUM_TOKENS = (int)size / 10;
	if (248 == (int)type)
	{
		tokcomp = 1;
		NUM_TOKENS /= 2;
	}
	else if (224 != (int)type)
	{
		DlgMessage("Invalid Input", "Token file variable type is not a text or appvar file", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		free(output);
		return;
	}
	
	//Alloc memory from token information
	if ((tokens = (tokenInfo *)calloc(NUM_TOKENS, sizeof(tokenInfo))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token info array", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(input);
		free(output);
		return;
	}
	if ((tokentext = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token data string", BT_OK, BT_NONE);
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(tokens);
		free(input);
		free(output);
		return;
	}
	
	//If the file is a binary file, get input directly, otherwise, if it's text, print dialog to say so and get text input
	if (tokcomp == 1)
		numtokens = readBinaryTokens(tokens, tokenfile, extension, NUM_TOKENS);
	else
	{
		DlgMessage("Solar89 Tokenizer", "The token file is a TEXT variable. Solar89 can parse these files, but it will be slower than using a TOKS variable instead.", BT_OK, BT_NONE);
		strcpy(extension, "8XP");
		ST_helpMsg("Initializing tokens");
		bool = getTextFromFile(tokenfile, tokentext, size);
		if (bool == 0)
		{
			free(string);
			free(inputfile);
			free(outputfile);
			free(tokenfile);
			free(tokens);
			free(input);
			free(output);
			free(tokentext);
			return;
		}
		
		//Initialize token infos and tokenize (functions in seperate file), if -1 return
		numtokens = initTokens(tokens, tokentext, NUM_TOKENS);
		if (numtokens == -1)
		{
			free(string);
			free(inputfile);
			free(outputfile);
			free(tokenfile);
			free(tokens);
			free(input);
			free(output);
			free(tokentext);
			return;
		}	
	}

	//Now, with the gathered token info objects, tokenize the text (add a boolean system here, so we exit before writing file if error?)
	ST_helpMsg("Tokenizing");
	tokenizePrgm(input, output, tokens, numtokens);

	//Write tokens to application variable
	bool = writePrgmToVar(outputfile, output, size, extension);
	if (bool == 0)
	{
		free(string);
		free(inputfile);
		free(outputfile);
		free(tokenfile);
		free(tokens);
		free(input);
		free(output);
		free(tokentext);
		return;
	}
	
	//Free memory
	free(string);
	free(inputfile);
	free(outputfile);
	free(tokenfile);
	free(tokens);
	free(input);
	free(output);
	free(tokentext);

	//Print message and wait for keypress
	ST_helpMsg("Process complete- press any key");
	ngetchx();
}
