#!/usr/bin/env python

"""make8xp [-s] [-2] [-3] [-5] [-6] [filename]

make8xp is a Python program designed to convert *.89y files (files from the 
TI-89) to program files for the TI-83+ and TI-84+ calculators, in addition 
to program files for the TI-82, TI-83, TI-85, and TI-86 calculators.

If you didn't understand any of that, you probably shouldn't be using this
program, and I suggest you either visit www.ticalc.org to find out more 
information about Texas Instruments' programmable calculators.

This program can be double-clicked or ran from the commandline, it makes no
difference. It makes use of various Python modules to extract hexadecimel
tokens from a text file and package them in a program file.

The -s switch specifies whether or not to save the resultant binary file,
consisting of only the program data itself. This is similar some assemblers
for z80 calculators.

The -2, -3, -5, and -6 switches allow you to tell the program to save as a
program file for the TI-82, TI-83, TI-85, or TI-86, instead of the default,
which is the TI-83+. (The TI-84+ and the silver editions of both have the 
same program file format). These switches are experimental and untested.
"""

import getopt
import os
import sys
from pySolar89 import parse89
from pySolar89 import binpac8x

def helpMsg():
	"""Print out the help text, then run the program."""
	print >> sys.stderr, __doc__
	raw_input()
	file = raw_input("File name: ")
	calc = raw_input("Calculator model [82/83/8x/85/86]: ")
	print ""
	convert(file, calc)

def convert(file, calc, save=False):
	"""Converts 89y variable file to 8xp (or 82p/83p/85p/86p) program file."""
	
	#Try to get the data from the *.89y file, if an IOError is raised, quit
	print "Retrieving data from file..."
	try:
		name, folder, type, data = parse89.open89var(file)
	except IOError:
		print "Failed to retrieve data from input file"
		print "The path and/or name of the file is likely incorrect"
		return
	print "Extracted data from file", name
	
	#Remove the tail end- the file type stuff as generated by the calculator software
	end = len(data) - 4
	data = data[:end]
	
	#Create a temporary binary file for binPac8x to use
	path, split, tail = file.partition('.')
	print "Writing binary data to file"
	binpath = path + '.bin'
	binfile = open(binpath, 'wb')
	binfile.write(data)
	binfile.close()

	#Call binPac8x functions to package the tokens as an *.8*p
	print "Running binPac8x to package binary file as program file"
	basepath = os.path.dirname(binpath)
	outpath = ""
	oncalcname = name.upper()
	binpac8x.parsefile(binpath, outpath, basepath, calc[1], oncalcname)
	
	#If user options say don't save, delete the binary file
	if save == False:
		print "Removing temporary binary file"
		os.remove(binpath)
		
def main():
	"""The script function, handles all commandline stuff."""
	save = False
	calc = "8x"
	
	#Try to get opts and args from the commandline
	try:
		opts, args = getopt.getopt(sys.argv[1:], "s2356", ["save", "82", "83", "85", "86"])
	except getopt.error, msg:
		pass
	
	#If we have no args, print help text before running- if we do, deal with them, and then run
	if ((len(opts) == 0) and (len(args) == 0)):
		helpMsg()
	else:
		for o, a in opts:
			if o in ('-s', '--save'):
				save = True
			elif o in ('-2', '--82'):
				calc = "82"
			elif o in ('-3', '--83'):
				calc = "83"
			elif o in ('-5', '--85'):
				calc = "83"
			elif o in ('-6', '--86'):
				calc = "83"
		for arg in args:
			convert(arg, calc, save)

	#Pause for keypress (to enable executable mode!)
	raw_input()

if __name__ == "__main__":
	main()