// C Source File
// Created 8/16/2010; 9:00:01 PM

// Token File Compiler for Solar89
// Currently stand-alone program
// Makes token files be twice as large, but SHOULD make them load much faster
//	That is untested!

// Delete or comment out the items you do not need.
#define COMMENT_STRING         	"Token File Compiler"
#define COMMENT_PROGRAM_NAME   	"Solar89 Token Compiler"
#define COMMENT_VERSION_NUMBER 	1,0,0,0
#define COMMENT_AUTHORS        	"TC01"
#define COMMENT_BW_ICON \
	{0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000, \
	 0b0000000000000000}

#include <tigcclib.h>

//Token info structure
typedef struct 
{
    size_t chars; 								// How many chars are in the string making up the token
    int twoByteToken; 						// Whether or not the token is two bytes (0 or 1)
    unsigned char hex[2]; 				// The two bytes making up the token, or only hex[0] if it's a one-byte token
    char name[14];								// The name of the token (never more than 14 chars!)
} tokenInfo;

//Token file structure
typedef struct
{
	int numtokens;									// The number of tokens
	char type[3];										// The type of file to produce
} tokenFile;

//Removes the first character of a string
void removeFirstChar(char *string)
{
	int i;
	int length = strlen(string) - 1;

	char backupchar;
	char *backup = string;

	//Loop over the characters of the pointer
	for (i = 0; i < length; ++i)
	{
		backupchar = backup[i+1];
		*(string++) = backupchar;
	}
	
	//Add a \0 in place of the last character
	*(string++) = '\0';
}

//Creates and appends a token info object to the tokens array
void createTokenInfo(tokenInfo *tokens, char *working, char *line, char *tokenstr, int i)
{
	int newlength;

	unsigned long rawhex;
	unsigned char hex = '0';
	
	tokenInfo token;

	//Init the token
	token = tokens[i];
	token.twoByteToken = 0;

	//If the line starts with a space, implement bugfix
	if (line[0] == ' ')
		removeFirstChar(line);

	//Seperate the name component from the line and set it (and it's length!)
	memset(working, '\0', strlen(working));
	sprintf(working, "%c", '\\');
	newlength = strcspn(line, working);
	tokenstr = strncat(tokenstr, line, newlength);
	line += (newlength + 1);
	token.chars = newlength;
	strncat(token.name, tokenstr, newlength);

	//Seperate the hex component from the line and set it
	memset(tokenstr, '\0', strlen(tokenstr));
	tokenstr = strncat(tokenstr, line, 4);
	rawhex = strtol(tokenstr, NULL, 0);
	hex = (unsigned char)rawhex;
	token.hex[0] = hex;
	line += 5;

	//Then check if what's left of the string is the second byte of a two-byte token and deal with it
	if (strcmp(line, "0") != 0)
	{
		rawhex = strtol(line, NULL, 0);
		hex = (unsigned char)rawhex;
		token.hex[1] = hex;
		token.twoByteToken = 1;
	}

	//Update the token
	tokens[i] = token;
}

//Initializes all token infos, return number of tokens
int initTokens(tokenInfo *tokens, char *tokentext, int NUM_TOKENS)
{
	int length;
	int span = 0;
	int i = 0;

	char newchar;
	char endline = '\r';

	char *working = NULL;
	char *line = NULL;
	char *tokenstr = NULL;

	if ((working = (char *)calloc(5, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for working string", BT_OK, BT_NONE);
		return -1;
	}
	if ((tokenstr = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token working string", BT_OK, BT_NONE);
		free(working);
		return -1;
	}
	if ((line = (char *)calloc(10, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for line string", BT_OK, BT_NONE);
		free(working);
		free(tokenstr);
		return -1;
	}

	//Get the length of the token file and realloc tokenstr (not sure if needed...)
	length = strlen(tokentext);

	//Loop until the length of the input string is zero
	while (length > 0)
	{
		//Reset the strings
		line = memset(line, '\0', strlen(line));
		tokenstr = memset(tokenstr, '\0', strlen(tokenstr));
		working = memset(working, '\0', strlen(working));

		//Get the length of the line up to trailing character, realloc mem
		sprintf(working, "%c", endline);
		span = strcspn(tokentext, working);
		if (tokenstr != NULL) 
			tokenstr = (char *)realloc(tokenstr, span * sizeof(char));
		if (line != NULL)
			line = (char *)realloc(line, span * sizeof(char));

		//Load the line into a string, and then remove it
		line = strncat(line, tokentext, span);
		tokentext = strpbrk(tokentext, working);
		
		//Then clear out the \r at the front of the string
		newchar = tokentext[2];
		sprintf(working, "%c", newchar);
		tokentext = strpbrk(tokentext, working);
		length = strlen(tokentext);

		//Then make a token info using some of these variables (to make this code easier to read, debug, and to fix a bug) and free the line memory
		createTokenInfo(tokens, working, line, tokenstr, i);
		i += 1;
	}
	
	//Free memory
	free(line);
	free(tokenstr);
	free(working);
	
	//Return the data
	return NUM_TOKENS;
}

//Makes a dialog to get input strings, returns boolean (integer) of 0 or 1 depending on success
int getDialogInput(char *string)
{
	int result = 0;
	
	HANDLE dlg = H_NULL;

	//Create a dialog box and add the necessary inputs
	if ((dlg = DialogNewSimple(140, 55)) == H_NULL) 
	{
		DlgMessage("DMA Failure", "Unable to create dialog input box", BT_OK, BT_CANCEL);
		return result;
	}
	DialogAddTitle(dlg, "Solar89 Token Parser", BT_OK, BT_CANCEL);
	DialogAddRequest(dlg, 5, 15, "Token Text File:", 0, 17, 18);
	DialogAddRequest(dlg, 5, 25, "Output File Name:", 17, 17, 18);
	DialogAddRequest(dlg, 30, 35, "File Type:", 34, 5, 5);
	
	//Prompt for input
	if (DialogDo(dlg, CENTER, CENTER, string, NULL) == KEY_ENTER)
		result = 1;

	HeapFree(dlg);
	
	//Return result to say we've done it
	return result;
}

//Extracts the text of a text file, returns a boolean value
int openTextFile(char *filename, char *input, unsigned short size)
{
	FILE *source = NULL;

	int result = 0;
	
	//Open the input file and get its text
	if ((source = fopen(filename, "rt")) == NULL)
	{
		DlgMessage("I/O Error", "Unable to read output file- this may be because the file name was entered incorrectly.", BT_OK, BT_NONE);
		return result;
	}
	
	result = 1;
	fread(input, size, 1, source);
	fclose(source);
	
	return result;
}

//Writes the token info object to a file, returns a boolean value
int writeTokenFile(char *filename, tokenFile tokfile, tokenInfo *tokens, int numtokens)
{
	FILE *output = NULL;
	int size;
	
	//Open a new file for writing in binary mode
	if ((output = fopen(filename, "wb")) == NULL)
	{
		DlgMessage("I/O Error", "Unable to create output file- this may be because that file already exists on your calc.", BT_OK, BT_NONE);
		return 0;
	}
	
	//Write the token file object
	fwrite(&tokfile, sizeof(tokfile), 1, output);

	//Write the token info array
	size = (unsigned short)numtokens * sizeof(tokenInfo);
	fwrite(tokens, size, 1, output);
	
	//Write the file type
	fputc(0, output);
	fputs("TOKS", output);
	fputc(0, output);
	fputc(OTH_TAG, output);

	//Close the file, archvie it, and return
	fclose(output);
	EM_moveSymToExtMem(SYMSTR(filename), HS_NULL);
	return 1;
}

void _main(void)
{
	int i = 0;
	int bool = 0;
	int argtype;
	int NUM_TOKENS = 0;
	int numtokens;
	int length;
	
	unsigned short size = 0;

	char *input = NULL;
	char *inputname = NULL;
	char *outputname = NULL;
	char *extension = NULL;
	char *tokentext = NULL;
	
	tokenInfo *tokens = NULL;
	tokenFile tokfile;
	SYM_ENTRY *sym;
	ESI argptr;
	
	//Alloc memory
	if ((input = (char *)calloc(60, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for raw input string", BT_OK, BT_NONE);
		return;
	}
	if ((inputname = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for input string", BT_OK, BT_NONE);
		return;
	}
	if ((outputname = (char *)calloc(20, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for output string", BT_OK, BT_NONE);
		return;
	}
	if ((extension = (char *)calloc(6, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for file extension string", BT_OK, BT_NONE);
		return;
	}
	
	//Check for arguments passed to the program
	if (ArgCount() == 3)
	{
		InitArgPtr(argptr);
		while ((argtype = GetArgType(argptr)) != END_TAG)
		{
			i += 1;
			if (argtype == STR_TAG)
			{
				switch (i)
				{
					case 1: 
						inputname = strcpy(inputname, GetStrnArg(argptr));
						continue;
					case 2: 
						outputname = strcpy(outputname, GetStrnArg(argptr));
						continue;
					case 3: 
						extension = strcpy(extension, GetStrnArg(argptr));
						continue;
					default: 
						continue;
				}
			}
			else
			{
				free(input);
				free(inputname);
				free(outputname);
				free(extension);
				ER_throw(ER_ARGUMENT);
			}
		}
	}
	//If there are NOT three arguments, get dialog input- if this is not possible, exit
	else
	{
		bool = getDialogInput(input);
		if (bool == 0)
		{
			free(input);
			free(inputname);
			free(outputname);
			free(extension);
			return;
		}
		
		//Retrieve file names from raw input
		for (length = 0; length < 17; ++length)
		{
			inputname[length] = input[length];
			outputname[length] = input[length + 17];
			if (length < 5)
				extension[length] = input[length + 34];
		}
	}

	//Now, get the size of the token info file and allocate mem
	ST_helpMsg("Reading input file");
	sym = SymFindPtr(SYMSTR(inputname), 0);
	size = ((MULTI_EXPR*)HeapDeref(sym->handle))->Size + 2;
	NUM_TOKENS = (int)size / 10;
	if ((tokentext = (char *)calloc(size, sizeof(char))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token data string", BT_OK, BT_NONE);
		free(input);
		free(inputname);
		free(outputname);
		free(extension);
		return;
	}
	
	//Read in data from token file- if this fails, exit and free memory
	bool = openTextFile(inputname, tokentext, size);
	if (bool == 0)
	{
		free(input);
		free(inputname);
		free(outputname);
		free(extension);
		free(tokentext);
		return;	
	}
	
	//Alloc memory for token info array
	if ((tokens = (tokenInfo *)calloc(NUM_TOKENS, sizeof(tokenInfo))) == NULL) 
	{
		DlgMessage("Alloc Failure", "Unable to allocate space for token info array", BT_OK, BT_NONE);
		free(input);
		free(inputname);
		free(outputname);
		free(extension);
		free(tokentext);
		return;
	}
	
	//Create a token info array from the token text, if -1 return
	ST_helpMsg("Parsing token file");
	numtokens = initTokens(tokens, tokentext, NUM_TOKENS);
	if (numtokens == -1)
	{
		free(input);
		free(inputname);
		free(outputname);
		free(extension);
		free(tokentext);
		free(tokens);
		return;
	}
	
	//Now, make a token file object (size uncertain... extension will be three bytes, int defining numtokens = two bytes... giving some breathing space
	ST_helpMsg("Writing token file");
	tokfile.numtokens = numtokens;
	strcpy(tokfile.type, extension);

	//Save as token file object- doesn't matter if bool = 0 or not because we're exiting anyway
	bool = writeTokenFile(outputname, tokfile, tokens, numtokens);
	
	//Free memory and exit!
	free(input);
	free(inputname);
	free(outputname);
	free(extension);
	free(tokentext);
	free(tokens);
}